#!/bin/bash -u

# Example script calls to produce masks for the GaBoDS
# image Deep1a_R.D1AA.swarp.fits and the associated 
# weight map Deep1a_R.D1AA.swarp.weight.fits

# NECESSARY SETUP TO RUN THIS SCRIPT:
# - Install the THELI pipeline 'theli-X.XX.XX.tgz'
# - Install the automask package 'automask-X.XX.XX.tar.gz'
# - create 'somewhere' a working directory. We call this
#   directory <MASKING> henceforth.
#   - Copy the file 'progs.ini' from ../theli-X.XX.XX/scripts/Linux
#     (or .../theli-X.XX.XX/scripts/Linux_64; it depends on 
#      your machine which one you have)
#     to <MASKING>
#   - Copy ALL files from .../automask/Linux(_64) to <MASKING>
#   - Copy this script to <MASKING>
# - Install the GSC catalogue 'GSC.tar.gz'
#   - After unpacking go to .../GSC/src/ and see whether the program
#     'gsc.exe' works; if not you have to compile it with the Makefile
#     in this directory; sorry I did not write down whether there were
#     problems when I did this. You have to make sure that 'gsc.exe'
#     finally ends up in .../GSC/bin !

# To call this script you need to copy it to <MASKING>
# I assume that automask scripts and progs.ini are now in the current
# directory; all images are in the dir. pointed to by the IMAGEDIR
# var. below. The GSC catalogue be in GSCDIR (see below)

IMAGEDIR=../images
GSCDIR=../GSC
AUTOMASK=../automask # where is 'automask' installed

# needed for the GSC1 utility program retrieving object data:
export GSCDAT=${GSCDIR}

# The image and weight to be analysed
ANALYSEIMA=Deep1a_R.D1AA.swarp.fits
ANALYSEWEIGHT=Deep1a_R.D1AA.swarp.weight.fits

# some sanity checks:
test -f ${IMAGEDIR}/${ANALYSEIMA} ||\
  { echo "Image ${IMAGEDIR}/${ANALYSEIMA} not present! Exiting!"; exit 1; }

test -f ${IMAGEDIR}/${ANALYSEWEIGHT} ||\
  { echo "Image ${IMAGEDIR}/${ANALYSEWEIGHT} not present! Exiting!"; exit 1; }

test -f ${GSCDIR}/bin/gsc.exe ||\
  { echo "File ${GSCDIR}/bin/gsc.exe not present! Exiting!"; exit 1; }

#
# start with main script work:
#
# To have a prefix name for created catalogues, masks etc.:
BASE=`basename ${ANALYSEIMA} .fits`

# In the following we create masks for stellar sources. We found that
# best results are obtained if using the GSC1 catalogue for bright and
# the GSC2 catalogur for faint stars. We implement this with the
# following two calls to 'maskstars_simple.sh'. Note also that the
# slope for the star size scaling with magnitude is different for the
# two samples (see also the documentation message of maskstars_simple.sh). 
 
# generate mask for 'bright' stars with GSC1; name of the output
# mask is '..brightstar.reg'
echo "Creating masks for bright stars"
./maskstars_simple.sh -i ${IMAGEDIR}/${ANALYSEIMA} \
                      -a GSC1 \
                      -d ${GSCDIR} \
                      -s ${AUTOMASK}/config/WFI_star_R_14.1.reg \
                      -p 14 -l 14 -m 0.9 \
                      -o ${BASE}_brightstar.reg

# generate mask for 'faint' stars with GSC2.2; name of the output
# mask is '..faintstar.reg'
echo "Creating masks for faint stars"
./maskstars_simple.sh -i ${IMAGEDIR}/${ANALYSEIMA} \
                      -a GSC2.2 \
                      -s ./${AUTOMASK}/config/WFI_star_R_14.1.reg \
                      -p 14 -l 17 -m 0.2 \
                      -o ${BASE}_faintstar.reg

# create 'object density' masks; they do not work well if the
# field-of-view is smaller than 0.25 sq. degree (one WFI field); in
# this case the maskvoids script typically performs a much too
# conservative masking:
echo "Creating masks for underdense and overdense regions"
./maskvoids.sh -i ${IMAGEDIR}/${ANALYSEIMA} \
               -w ${IMAGEDIR}/${ANALYSEWEIGHT} \
               -r ${BASE}_voids.reg -d 8

# In the following we want to mask objects which contain pixels of
# weight zero and which have not been masked by the steps done
# hitherto; see also comments before the 'create_flaggedstars_masks.sh' 
# command; very probably you need to heavily modify this part of
# the masking if your data were not processed by THELI.
#
# First, paste the masks created hitherto; if masking zero-weight objects
# we want to exclude objects which have a mask already.
cat ${BASE}_brightstar.reg ${BASE}_faintstar.reg ${BASE}_voids.reg > \
  ${BASE}_star_voids.reg

# create a flag map with all saturated (weight zero) pixels set to '1'; 
# all other pixels set to zero. The script 'create_flaggedstars_masks.sh' 
# below then masks objects which are flagged but which are not yet masked:
echo "creating Flag map ${IMAGEDIR}/${BASE}.flag.fits"
ic -p 8 '1 0 %1 fabs 1.0e-06 < ?' ${IMAGEDIR}/${ANALYSEWEIGHT} > \
  ${IMAGEDIR}/${BASE}.flag.fits

# This masking step does not produce new masks for the image
# Deep1a_R.D1AA.swarp.fits:
echo "flagging remaining saturated/zero weight features"
./create_flaggedstars_masks.sh ${IMAGEDIR}\
  ${ANALYSEIMA} ${ANALYSEWEIGHT} ${BASE}.flag.fits \
  WFI \
  ./${BASE}_star_voids.reg \
  ./${BASE}_saturated.reg

FINALMASKS="${BASE}_star_voids.reg"

if [ -f ./${BASE}_saturated.reg ]; then
  FINALMASKS="${FINALMASKS} ./${BASE}_saturated.reg"  
fi

# finally paste all obtained mask to a 'final' one:
cat ${FINALMASKS} > ${BASE}_final.reg

# clean up and bye
rm ./${BASE}_star_voids.reg

exit 0;
