#!/usr/bin/env perl

# This script searches for objects in a row which could be
# asteroid trails and masks them

use constant PI    => 4 * atan2(1, 1);

$catalog=$ARGV[0]; # Catalog with entries Xpos Ypos A B THETA
$boxsize=$ARGV[1]; # objects are considered in boxes (~100) 
$nboxestogether=$ARGV[2]; # to ensure that tracks between boxes are found
                          # always rejoin this number^2 of boxes, and keep
                          # 1 box overlap (~5)
                          # must be odd.
$transverseTolerance=$ARGV[3]; # allowed offset from line
$minNumDetections=$ARGV[4]; # min. number to call an asteroid # if =0: take optimal number 
$widthmask=$ARGV[5]; # width of mask
$extensionmask=$ARGV[6]; # extra extension of mask in asteroid-direction
$emedian_thresh=$ARGV[7]; # Abs(Median ellipticity) threshold to consider this an asteroid 
$max_spurious_fraction=$ARGV[8]; #Maximum accepted fraction of spurious detections. If the measured
                                 # fraction is higher, the programme will repeat with $minNumDetections++;

$min_join_region=$minNumDetections-2; # min number of common matches in the current and a previous positiv
                    # detection to consider that this is the same object

print "\n*** mask_asteroids.pl $catalog $boxsize $nboxestogether $transverseTolerance $minNumDetections $widthmask $extensionmask $emedian_thresh $max_spurious_fraction ***\n\n";

unless ($nboxestogether % 2) {
    die "error: Number of joint boxes must be odd. You chose $nboxestogether\n";
}

$maxsep2=$transverseTolerance*$transverseTolerance;

unless($max_spurious_fraction){$max_spurious_fraction=1};

print "Catalog: $catalog\n";
print "Boxsize: $boxsize\n";
print "Number of joint boxes: $nboxestogether * $nboxestogether\n";
print "Transverse tolerance: $transverseTolerance\n";
print "Min Number of detections: $minNumDetections\n";
print "Width of mask: $widthmask\n";
print "Mask extension factor: $extensionmask\n";
print "Abs(Median ellipticity) threshold to consider a detection an asteroid: $emedian_thresh\n";
print "Maximum accepted fraction of spurious detections: $max_spurious_fraction. \n";

print "Read catalog...\n";
if (-r ($catalog)) {
	open (DATAFILE,($catalog));  
	$catBuf  = join("", <DATAFILE>);      
	close (DATAFILE);
} else {
	die "File $catalog not found!";
}

print "Setup grid...\n";

$catBuf=~s/^\s*//igs;
$catBuf=~s/\s*$//igs;

%binEntriesHash;
%objectHash;

$x_min=10000000;
$y_min=10000000;
$x_max=0;
$y_max=0;

$nobj=0;
foreach (split /\n/, $catBuf) {
    $line=$_;
    $line=~s/^\s*//igs;
    $line=~s/\s*$//igs;
    ($x,$y,@rest)=split /\s\s*/, $line;
    $xbin=int($x/$boxsize);
    $ybin=int($y/$boxsize);
    $key=$xbin.'_'.$ybin;
    $objectkey=$x.'_'.$y;
    $binEntriesHash{$key}.=$objectkey."\n";
    $objectHash{$objectkey}=$line;
    if ($x<$x_min){$x_min=$x;}
    if ($y<$y_min){$y_min=$y;}
    if ($x>$x_max){$x_max=$x;}
    if ($y>$y_max){$y_max=$y;}
    $nobj++;
}
print "x: $x_min .. $x_max\n";
print "y: $y_min .. $y_max\n";

print "Nobject: $nobj\n";
# Determine grid size:

$xbin_min=10000;
$ybin_min=10000;
$xbin_max=0;
$ybin_max=0;
$nbins=0;
foreach (keys %binEntriesHash) {
    $key=$_;
    ($xbin,$ybin,@rest)=split /\_/, $key;
    if ($xbin<$xbin_min){$xbin_min=$xbin;}
    if ($ybin<$ybin_min){$ybin_min=$ybin;}
    if ($xbin>$xbin_max){$xbin_max=$xbin;}
    if ($ybin>$ybin_max){$ybin_max=$ybin;}
    $nbins++;
}
print "Nbins: $nbins\n";

$ObjectDensity=$nobj/(($x_max-$x_min)*($y_max-$y_min));

print "Object density: $ObjectDensity /pixel^2\n";

print "xbins: $xbin_min .. $xbin_max\n";
print "ybins: $ybin_min .. $ybin_max\n";

print "rejoin grid...\n";

$steps_from_bin=int($nboxestogether/2);

%rebinnedHash;
$count=1;
for ($xbin=$xbin_min;$xbin<=($xbin_max+$steps_from_bin);$xbin+=$steps_from_bin*2){
    for ($ybin=$ybin_min;$ybin<=($ybin_max+$steps_from_bin);$ybin+=$steps_from_bin*2){
#	print "rejoin around bin center xbin=$xbin, ybin=$ybin\n";
	$joinBuf='';
	for($i=-$steps_from_bin;$i<=$steps_from_bin;$i++)
	{
	    for($j=-$steps_from_bin;$j<=$steps_from_bin;$j++)
	    {
		$newxbin=$xbin+$i;
		$newybin=$ybin+$j;
		$key=$newxbin.'_'.$newybin;
		$joinBuf.=$binEntriesHash{$key};
	    }
	}
	$rebinnedHash{$count}=$joinBuf;
	$count++;
    }
}

$iterate=1;
$iterate_counter=1;

while ($iterate) {

@allbin_foundList=();
$count2=1;
$allbin_newkey_counter=0;
print "Scan for alignment: Iteration $iterate_counter ...\n";
foreach (keys %rebinnedHash) {
    $key=$_;
    $val=$rebinnedHash{$key};
    # start a seperate found list for each bin
    @foundList=();
    $newkey_counter=0;


    if ($val) {
#	print "analyse new bin center: $key ($count2 / $count)\n";
	$val=~s/\n$//is;
	(@objects)=split /\n/, $val;
	$nobj=$#objects+1;
#	print "Number objects: $nobj\n";
	# 2 loops to define all pairs from which to define a line
	for ($i=0;$i<$nobj;$i++){
	    ($x1,$y1,@rest)=split /_/, @objects[$i];
	    for ($j=$i+1;$j<$nobj;$j++){
		($x2,$y2,@rest)=split /_/, @objects[$j];
		# P1=i, P2=j, use P1 as reference. 
		# Compute normal vector of line defined by the two points:
		$nx=($y1-$y2);
		$ny=($x2-$x1);
		$nnorm=sqrt($nx*$nx+$ny*$ny);
		$nx/=$nnorm;
		$ny/=$nnorm;
		$matches=2;
		$currentMatchList=@objects[$i]."\n".@objects[$j];
		# now a third loop, to compare the other stars with that line
		for ($k=0;$k<$nobj;$k++){
		    ($x3,$y3,@rest)=split /_/, @objects[$k];
		    unless (($k==$i)||($k==$j)) {
			# compute seperation from line
			$sep=($x3-$x1)*$nx+($y3-$y1)*$ny;
			$sep2=$sep*$sep;
			if ($sep2<$maxsep2) {
			    # We have a positive match!!
			    $matches++;
			    $currentMatchList.="\n".@objects[$k];
			}
		    }

		}
		if ($matches>=$minNumDetections) {
#		    print "$matches matches for x1=$x1, y1=$y1  x2=$x2, y2=$y2\n";
		    if ($newkey_counter) {
			# check if we have already found this astroid
			$astroidKnown=0;
#			foreach (@foundInThisBin)
#			{    
#			$PreviousEntryThisBin=$_;
			for($fcount=0;$fcount<$newkey_counter;$fcount++){
			    $identicalCounter=0;
			    $addList='';
			    foreach (split /\n/, $currentMatchList){
				$newObject=$_;
				$objectCrossIdentified=0;
				foreach (split /\n/, @foundList[$fcount]){
				    $oldObject=$_;
				    if ($newObject eq $oldObject) {
#					print "same objects: $newObject $oldObject\n";
					$identicalCounter++;
					$objectCrossIdentified=1;
				    }
				}
				unless ($objectCrossIdentified) {
				    $addList.="\n".$newObject;
				}
			    }
			    if ($identicalCounter >= $min_join_region) {
#				print "found $identicalCounter common objects\n";
				@foundList[$fcount].=$addList;
#				print "This looks like previous astroid number $fcount -> Merge. newkey_counter=$newkey_counter\n";
				$astroidKnown=1;
			    } 
			}
			unless($astroidKnown){
			    @foundList[$newkey_counter]=$currentMatchList;
#			    print "Candidate with $matches matches => number $newkey_counter (bin $count2 of $count)\n";
			    $newkey_counter++;				
			}
		    } else {
			# create the entry in the asteroid list
			@foundList[$newkey_counter]=$currentMatchList;
#			print "$allbin_newkey_counter : first asteroid  with $matches matches in bin $count2 of $count\n";
			$newkey_counter++;
		    }
		}

	    }
	}
    }
    $count2++;
    unless ($count2 % 1000) {print "$count2 of $count bins done -- $allbin_newkey_counter candidates\n";}
    $allbin_newkey_counter+=$newkey_counter;
    foreach (@foundList) {
	push(@allbin_foundList,$_);
    }
}
print "Found a total of $allbin_newkey_counter asteroid candidates\n";
print "Filter out probable galaxy chance alignments...\n";

$regionBuf='# Region file format: DS9 version 3.0
# Auto created by mask_asteroids.pl (Tim Schrabback, Leiden)
# Filename: masked.fits
global color=red font="helvetica 10 normal" select=1 edit=1 move=1 delete=1 include=1 fixed=0 source
';
$regionBuf_chance='# Region file format: DS9 version 3.0
# Auto created by mask_asteroids.pl (Tim Schrabback, Leiden)
# Filename: masked.fits
global color=blue font="helvetica 10 normal" select=1 edit=1 move=1 delete=1 include=1 fixed=0 source
';
$countaccept=0;
$countreject=0;

# Now fit the positions to get a precise line and create the mask
for ($i=0; $i<$allbin_newkey_counter; $i++){
    # Make a least square fit:
    $Sab=0;
    $Saa=0;
    $Sa=0;
    $Sb=0;
    $Sn=0;
    # If we have diffraction spikes or similar stuff the line can be very steep
    # Solution: check if data varies more in Xpos or Ypos direction and use this
    # as x value (a) in the fit
    $x_min=10000000;
    $y_min=10000000;
    $x_max=0;
    $y_max=0;
     foreach (split /\n/, @allbin_foundList[$i]){
	($x,$y,@rest)=split /\_/, $_;
	if ($x<$x_min){$x_min=$x;}
	if ($y<$y_min){$y_min=$y;}
	if ($x>$x_max){$x_max=$x;}
	if ($y>$y_max){$y_max=$y;}
    }
    if (($x_max-$x_min)>($y_max-$y_min)){
	$aISx=1;
	$a_min=$x_min;
	$a_max=$x_max;
    } else {
	$aISx=0;
	$a_min=$y_min;
	$a_max=$y_max;
    }
 
    $a_min_mask=($a_min+$a_max)/2-($a_max-$a_min)*$extensionmask/2;
    $a_max_mask=($a_min+$a_max)/2+($a_max-$a_min)*$extensionmask/2;
    foreach (split /\n/, @allbin_foundList[$i]){
	($x,$y,@rest)=split /\_/, $_;
	if ($aISx) {
	    $a=$x;
	    $b=$y;
	} else {
	    $a=$y;
	    $b=$x;
	}
	$Sab+=$a*$b;
	$Saa+=$a*$a;
	$Sa+=$a;
	$Sb+=$b;
	$Sn++;
    }
    $slope=($Sab-$Sa*$Sb/$Sn)/($Saa-$Sa*$Sa/$Sn);
    $axis_off=$Sb/$Sn-$slope*$Sa/$Sn;
    
    # start and end point of line:

    $al_min=$a_min_mask;
    $al_max=$a_max_mask;
    $bl_min=$a_min_mask*$slope+$axis_off;
    $bl_max=$a_max_mask*$slope+$axis_off;
#   print "aIsx $aISx $al_min $al_max $bl_min $bl_max\n";

    # Now compute da and db offsets due to rectangular mask
    $cosphi=sqrt(1/(1+$slope*$slope));
    $sinphi=$cosphi*$slope;

    $da=$widthmask*$sinphi/2;
    $db=$widthmask*$cosphi/2;

    $a1=$al_min+$da;
    $b1=$bl_min-$db;

    $a2=$al_min-$da;
    $b2=$bl_min+$db;

    $a3=$al_max-$da;
    $b3=$bl_max+$db;

    $a4=$al_max+$da;
    $b4=$bl_max-$db;
    if ($aISx) {
	$outLine='polygon('.$a1.','.$b1.','.$a2.','.$b2.','.$a3.','.$b3.','.$a4.','.$b4.')';
    } else {
	$outLine='polygon('.$b1.','.$a1.','.$b2.','.$a2.','.$b3.','.$a3.','.$b4.','.$a4.')'; 
    }

    # Check if this is likely a chance alignment of galaxies or truely an asteroid 
    # Use median ellipticity...
    @e1Array=();
    @e2Array=();
     foreach (split /\n/, @allbin_foundList[$i]){
	($x,$y,@rest)=split /\_/, $_;
	$objectkey=$x.'_'.$y;
	(@data)=split /\s\s*/, $objectHash{$objectkey};
	$A=@data[2];
 	$B=@data[3];
 	$theta=@data[4];
	$e1=(($A-$B)/($A+$B))*cos((2*$theta)*(PI/180.));
	$e2=(($A-$B)/($A+$B))*sin((2*$theta)*(PI/180.));
	push(@e1Array,$e1);
	push(@e2Array,$e2);
    }
    # Now compute the median ellipticity
    @Array = sort(NummernSort @e1Array);
    $even=$#Array%2; # $#Array = length -1
    if ($even) {
	$e1median = ($Array[int($#Array / 2)]+$Array[int($#Array / 2)+1])/2;
    } else {
	$e1median = $Array[int($#Array / 2)];
    }
    @Array = sort(NummernSort @e2Array);
    $even=$#Array%2; # $#Array = length -1
    if ($even) {
	$e2median = ($Array[int($#Array / 2)]+$Array[int($#Array / 2)+1])/2;
    } else {
	$e2median = $Array[int($#Array / 2)];
    }
    $emedian=sqrt($e1median*$e1median+$e2median*$e2median);
    if ($emedian<$emedian_thresh) {
#	print "emd=$emedian   --  this seems to be a chance alignment\n";
#	$regionBuf_chance.=$outLine."\n";
	$regionBuf_chance.=$outLine.'#em '.$emedian."\n";
	$countreject++;
    } else {
#	print "emd=$emedian   --  this seems to be a real asteroid\n";
#	$regionBuf.=$outLine.'#em '.$emedian."\n";
	$regionBuf.=$outLine."\n";
	$countaccept++;
    }

}
print "|median(e)|<$emedian_thresh => $countaccept asteroids, $countreject spurious\n";

$outFileadd='';
if ($countaccept) {
    $spurious_fraction=$countreject/($countreject+$countaccept);
    if ($spurious_fraction>$max_spurious_fraction){
	print "WARNING: spurious fraction $spurious_fraction larger than limit $max_spurious_fraction. Re-iterate...\n";
	$minNumDetections++;
	print "Min Number of detections now: $minNumDetections\n";
	$outFileadd='.iter'.$iterate_counter;
    } else {
	$iterate=0;
    }
    $iterate_counter++;
} else {
    print "No accepted candidate. -> Stop iteration.\n";
    $iterate=0;
}

$outFile='>'.$catalog.'.astmask'.$outFileadd.'.reg';
open NEWFILE, "$outFile";
print NEWFILE "$regionBuf";
close NEWFILE;

print "Wrote regions to $outFile\n";

$outFile='>'.$catalog.'.astmask_chance'.$outFileadd.'.reg';
open NEWFILE, "$outFile";
print NEWFILE "$regionBuf_chance";
close NEWFILE;

print "Wrote regions to $outFile\n";

# while ... end
}

sub NummernSort {
 if($a < $b)
  { return -1; }
 elsif($a == $b)
  { return 0; }
 else
  { return 1; }
}
